// $Id: Set.hh 1.4 Fri, 18 Jul 1997 15:53:23 -0700 wlee $
// 
//  Copyright =(c) 1994 by the University of Southern California
//  and/or the International Business Machines Corporation.
//  All rights reserved.
//
//  Permission to use, copy, modify, and distribute this software and
//  its documentation in source and binary forms for lawful
//  non-commercial purposes and without fee is hereby granted, provided
//  that the above copyright notice appear in all copies and that both
//  the copyright notice and this permission notice appear in supporting
//  documentation, and that any documentation, advertising materials,
//  and other materials related to such distribution and use acknowledge
//  that the software was developed by the University of Southern
//  California, Information Sciences Institute and/or the International
//  Business Machines Corporation.  The name of the USC or IBM may not
//  be used to endorse or promote products derived from this software
//  without specific prior written permission.
//
//  NEITHER THE UNIVERSITY OF SOUTHERN CALIFORNIA NOR INTERNATIONAL
//  BUSINESS MACHINES CORPORATION MAKES ANY REPRESENTATIONS ABOUT
//  THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS
//  PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, TITLE, AND 
//  NON-INFRINGEMENT.
//
//  IN NO EVENT SHALL USC, IBM, OR ANY OTHER CONTRIBUTOR BE LIABLE FOR ANY
//  SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, WHETHER IN CONTRACT,
//  TORT, OR OTHER FORM OF ACTION, ARISING OUT OF OR IN CONNECTION WITH,
//  THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
//  Questions concerning this software should be directed to 
//  info-ra@isi.edu.
//
//  Author(s): Cengiz Alaettinoglu <cengiz@isi.edu>

#ifndef SET_H
#define SET_H

#include "config.hh"
#include <ostream.h>
extern "C" {
#include <sys/types.h>
}

#include "Pix.h"
#include "debug.hh"

// base type for SetOfPix and SetOfSymbol
// needed so that NormalTerm class can uniformly access them
class Set {
public:
   virtual int empty() = 0;
   virtual int universal() = 0;
   virtual void make_universal() = 0;
   virtual void clear() = 0;
   virtual int add(Pix item) = 0; // add item; return non-zero if success

   virtual void operator |= (Set& b) = 0; // union
   virtual void operator &= (Set& b) = 0; // intersection
   virtual void operator ~  ()       = 0; // complement
   virtual int  operator == (Set& b) = 0; // equivalance
   virtual void operator =  (Set& b) = 0; // assignment
   friend ostream& operator<<(ostream& stream, Set& nt);
   virtual void do_print (ostream& stream) = 0;

   CLASS_DEBUG_MEMORY_HH(Set);

   virtual ~Set() {}
};

#endif   // SET_H
