// $Id: route-list.hh 1.3.2.1 Mon, 15 Dec 1997 14:43:04 -0800 cengiz $
// 
//  Copyright (c) 1994 by the University of Southern California
//  and/or the International Business Machines Corporation.
//  All rights reserved.
//
//  Permission to use, copy, modify, and distribute this software and
//  its documentation in source and binary forms for lawful
//  non-commercial purposes and without fee is hereby granted, provided
//  that the above copyright notice appear in all copies and that both
//  the copyright notice and this permission notice appear in supporting
//  documentation, and that any documentation, advertising materials,
//  and other materials related to such distribution and use acknowledge
//  that the software was developed by the University of Southern
//  California, Information Sciences Institute and/or the International
//  Business Machines Corporation.  The name of the USC or IBM may not
//  be used to endorse or promote products derived from this software
//  without specific prior written permission.
//
//  NEITHER THE UNIVERSITY OF SOUTHERN CALIFORNIA NOR INTERNATIONAL
//  BUSINESS MACHINES CORPORATION MAKES ANY REPRESENTATIONS ABOUT
//  THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS
//  PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, TITLE, AND 
//  NON-INFRINGEMENT.
//
//  IN NO EVENT SHALL USC, IBM, OR ANY OTHER CONTRIBUTOR BE LIABLE FOR ANY
//  SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, WHETHER IN CONTRACT,
//  TORT, OR OTHER FORM OF ACTION, ARISING OUT OF OR IN CONNECTION WITH,
//  THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
//  Questions concerning this software should be directed to 
//  info-ra@isi.edu.
//
//  Author(s): Cengiz Alaettinoglu <cengiz@isi.edu>

#ifndef ROUTE_LIST_H
#define ROUTE_LIST_H

#include "config.hh"
#include <cstring>
#include "List.hh"
#include "_SetOfPix.hh"
#include "roe_tcl.hh"
#include "RadixSet.hh"

class RouteList {
public:
   class Route {
   private:
      class db_as {
      public:
	 db_as() : db_as_l(this) {}
	 db_as(char *_db, char *_as) : db_as_l(this) {
	    strncpy(db, _db, sizeof(db));
	    strncpy(as, _as, sizeof(as));
	 }
	 char db[64];
	 char as[16];
	 
	 ListNode db_as_l;
      };

   public:
      Pix    route;
      
      int    display_index;
      int    indentation;

      char   reg_status;
#define VR_REGISTERED_UNDTRMND  0x01
#define VR_REGISTERED_NO        0x02
#define VR_REGISTERED_SINGLY    0x04
#define VR_REGISTERED_MULTI     0x08
#define VR_REGISTERED_YES       (VR_REGISTERED_SINGLY | VR_REGISTERED_MULTI)
      char   rtd_status;
#define VR_ROUTED_UNDTRMND      0x01
#define VR_ROUTED_YES       	0x02
#define VR_ROUTED_NO        	0x04
      char   hmd_status;
#define VR_HOMED_UNDTRMND       0x01
#define VR_HOMED_ME         	0x02
#define VR_HOMED_OTHER      	0x04
#define VR_HOMED_MULTI_ME      	0x08
#define VR_HOMED_MULTI_OTHER   	0x10
      char   status;
#define VR_IRR_QUERIED          0x01
#define VR_IRR_DONT_QUERY       0x02
      ListNode rlist;
      ListHead<db_as> db_as_l;

      Route()      : route(NULL), rlist(this) {
	 display_index = -1;
	 indentation = 0; 

	 status     = 0;
	 reg_status = VR_REGISTERED_UNDTRMND;
	 rtd_status = VR_ROUTED_UNDTRMND;
	 hmd_status = VR_HOMED_UNDTRMND;
      }

      Route(Pix i) : route(i), rlist(this) {
	 display_index = -1;
	 indentation = 0; 

	 status     = 0;
	 reg_status = VR_REGISTERED_UNDTRMND;
	 rtd_status = VR_ROUTED_UNDTRMND;
	 hmd_status = VR_HOMED_UNDTRMND;
      }
   };

   static void get_registrations(ClientData clientdata);

private:
   static RouteList::Route *get_registration_response();
   static RouteList::Route *ask_registration();
   void get_registration_now(RouteList::Route *vr);
   static void parse_registration_response(char *response,RouteList::Route *vr);

   ListHead<Route> rlist;
   Radix<Route>    radix;
   Route **vrdisplayed;
   int vrdisplayed_size;

   _SetOfPix registered_routes;
   _SetOfPix real_routes;

   void get_registrations_init(Route *first);
   void get_more_specifics_add(RadixNode *rdnode);
   void display_i(RadixNode *rdnode);

   char *format(Route *vr);

public:
   RouteList() : irr_radix(NULL) { init(); }

   char as[16];

   void display();
   void select(int index);

   void init(void);
   void load(char *_as);
   void read_bgp_dump(char *fname);
   void get_more_specifics(char *selection);
   void get_less_specifics(char *selection);

   Route *operator() (int index) {
      return vrdisplayed[index];
   }

   // Added by wlee@isi.edu
   Route *get_response_next(void) { return vr_response_next; }

private:
   int index;
   int indent;
   Route *vr_query_next;
   Route *vr_response_next;
   Radix<Route> *irr_radix;

   void define_irr_radix();   

   int matching_reg;
   int matching_rtd;
   int matching_hmd;
};

extern RouteList routelist;

#endif   // ROUTE_LIST_H
